from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import DamageIncreaseValue

from ...summon.base import AttackerSummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class GardenOfPurity_3_6(AttackerSummonBase):
    name: Literal["Garden of Purity"] = "Garden of Purity"
    version: Literal["3.6"] = "3.6"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 2

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            assert mode == "REAL"
            value.damage += 1
        return value


# Skills


class KamisatoArtKyouka(ElementalSkillBase):
    name: Literal["Kamisato Art: Kyouka"] = "Kamisato Art: Kyouka"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(
            match,
            [
                self.create_character_status(
                    "Takimeguri Kanka", {"usage": 3, "max_usage": 3}
                ),
            ],
        )


class KamisatoArtSuiyuu(ElementalBurstBase):
    name: Literal["Kamisato Art: Suiyuu"] = "Kamisato Art: Suiyuu"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match, [self.create_summon("Garden of Purity")])


# Talents


class KyoukaFuushi_3_6(SkillTalent):
    name: Literal["Kyouka Fuushi"]
    version: Literal["3.6"] = "3.6"
    character_name: Literal["Kamisato Ayato"] = "Kamisato Ayato"
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)
    skill: Literal["Kamisato Art: Kyouka"] = "Kamisato Art: Kyouka"


# character base


class KamisatoAyato_4_1(CharacterBase):
    name: Literal["Kamisato Ayato"]
    version: Literal["4.1"] = "4.1"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | KamisatoArtKyouka | KamisatoArtSuiyuu] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Kamisato Art: Marobashi",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            KamisatoArtKyouka(),
            KamisatoArtSuiyuu(),
        ]


register_class(KamisatoAyato_4_1 | KyoukaFuushi_3_6 | GardenOfPurity_3_6)



from typing import Literal

from ...utils.class_registry import register_class

from ..consts import DamageElementalType
from .base import AttackerSummonBase


class CryoHilichurlShooter_3_3(AttackerSummonBase):
    name: Literal["Cryo Hilichurl Shooter"] = "Cryo Hilichurl Shooter"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1


class ElectroHilichurlShooter_3_3(AttackerSummonBase):
    name: Literal["Electro Hilichurl Shooter"] = "Electro Hilichurl Shooter"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1


class HilichurlBerserker_3_3(AttackerSummonBase):
    name: Literal["Hilichurl Berserker"] = "Hilichurl Berserker"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1


class HydroSamachurl_3_3(AttackerSummonBase):
    name: Literal["Hydro Samachurl"] = "Hydro Samachurl"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 1


register_class(
    CryoHilichurlShooter_3_3
    | ElectroHilichurlShooter_3_3
    | HilichurlBerserker_3_3
    | HydroSamachurl_3_3
)



from typing import Any, Literal, List

from ....utils.class_registry import register_class

from ...action import RemoveObjectAction

from ...event import MakeDamageEventArguments

from ...consts import ElementType, IconType, SkillType

from ...modifiable_values import DamageIncreaseValue
from .base import RoundCharacterStatus


class HeavyStrike_3_7(RoundCharacterStatus):
    name: Literal["Heavy Strike"] = "Heavy Strike"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        Increase damage for normal attack by 1 and decrease usage.
        If it is charged attack, increase damage more 1.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        if self.usage <= 0:
            # no usage, not modify
            return value
        # this character use normal attack, increase damage by 1
        value.damage += 1
        # if it is charged attack, increase damage more 1
        if match.player_tables[self.position.player_idx].charge_satisfied:
            value.damage += 1
        # decrease usage
        self.usage -= 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class ShatteringIce_3_3(RoundCharacterStatus):
    name: Literal[
        "Elemental Resonance: Shattering Ice"
    ] = "Elemental Resonance: Shattering Ice"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        add 2 damage for skills.
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not this character use skill, do nothing
            return value
        if self.usage <= 0:
            # no usage, do nothing
            return value
        # we trigger elemental reaction, add 2 damage
        assert mode == "REAL"
        self.usage -= 1
        value.damage += 2
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When make damage end, check whether to remove.
        """
        return self.check_should_remove()


class FerventFlames_3_3(RoundCharacterStatus):
    name: Literal[
        "Elemental Resonance: Fervent Flames"
    ] = "Elemental Resonance: Fervent Flames"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If we trigger pyro elemental reaction, add 3 damage.
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None, False
        ):
            # not this character use skill, do nothing
            return value
        if ElementType.PYRO not in value.reacted_elements:
            # not trigger pyro elemental reaction, do nothing
            return value
        if self.usage <= 0:
            # no usage, do nothing
            return value
        # we trigger elemental reaction, add 2 damage
        assert mode == "REAL"
        self.usage -= 1
        value.damage += 3
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When make damage end, check whether to remove.
        """
        return self.check_should_remove()


register_class(HeavyStrike_3_7 | ShatteringIce_3_3 | FerventFlames_3_3)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import ObjectPositionType, SkillType

from ....action import ChargeAction

from ....event import SkillEndEventArguments

from ....struct import Cost
from .base import RoundEffectArtifactBase


class ExilesCirclet_3_3(RoundEffectArtifactBase):
    name: Literal["Exile's Circlet"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(any_dice_number=2)
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChargeAction]:
        """
        If self use elemental burst, charge standby characters.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            source_area=ObjectPositionType.CHARACTER,
            target_area=ObjectPositionType.SKILL,
        ):
            # not equipped, or not this character use skill
            return []
        if self.usage <= 0:
            # no usage left
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_BURST:
            # not elemental burst
            return []
        # charge
        self.usage -= 1
        ret: List[ChargeAction] = []
        for cid, character in enumerate(
            match.player_tables[self.position.player_idx].characters
        ):
            if cid == self.position.character_idx or character.is_defeated:
                # skip self and defeated characters
                continue
            ret.append(
                ChargeAction(
                    player_idx=self.position.player_idx,
                    character_idx=cid,
                    charge=1,
                )
            )
        return ret


register_class(ExilesCirclet_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import CreateObjectEventArguments, RoundPrepareEventArguments

from ...summon.base import AttackerSummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class CuileinAnbar_3_3(AttackerSummonBase):
    name: Literal["Cuilein-Anbar"] = "Cuilein-Anbar"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 2


# Skills


class FloralBrush(ElementalSkillBase):
    name: Literal["Floral Brush"] = "Floral Brush"
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=3,
    )
    create_status_counter: int = 0

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        In prepare, reset counter
        """
        self.create_status_counter = 0
        return []

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        If our character create Floral Sidewinder, increase create status
        """
        if event.action.object_position.player_idx != self.position.player_idx:
            # not our character
            return []
        if event.action.object_name != "Floral Sidewinder":
            # not Floral Sidewinder
            return []
        self.create_status_counter += 1
        return []

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If talent equipped, before attack, create status.
        """
        ret = super().get_actions(match)
        if self.create_status_counter > 0:
            # created this round, not create
            return ret
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.talent is None:
            # no talent equipped
            return ret
        return super().get_actions(
            match, [self.create_team_status("Floral Sidewinder")]
        )


class TrumpCardKitty(ElementalBurstBase):
    name: Literal["Trump-Card Kitty"] = "Trump-Card Kitty"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=3,
        charge=2,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match, [self.create_summon("Cuilein-Anbar")])


# Talents


class FloralSidewinder_3_4(SkillTalent):
    """
    Because before using skill, the status should appear, and status need to
    listen DAMAGE_RECEIVE event, we should modify both actions of
    elemental skill and equipping this card.
    """

    name: Literal["Floral Sidewinder"]
    version: Literal["3.4"] = "3.4"
    character_name: Literal["Collei"] = "Collei"
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=4)
    skill: Literal["Floral Brush"] = "Floral Brush"


class FloralSidewinder_3_3(FloralSidewinder_3_4):
    version: Literal["3.3"]
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)


# character base


class Collei_3_3(CharacterBase):
    name: Literal["Collei"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | FloralBrush | TrumpCardKitty] = []
    faction: List[FactionType] = [FactionType.SUMERU]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Supplicant's Bowmanship",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            FloralBrush(),
            TrumpCardKitty(),
        ]


register_class(
    Collei_3_3 | FloralSidewinder_3_3 | FloralSidewinder_3_4 | CuileinAnbar_3_3
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackAndGenerateStatusSummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class FierySanctumField_4_1(AttackAndGenerateStatusSummonBase):
    name: Literal["Fiery Sanctum Field"] = "Fiery Sanctum Field"
    version: Literal["4.1"] = "4.1"
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1


# Skills


class MoltenInferno(ElementalSkillBase):
    name: Literal["Molten Inferno"] = "Molten Inferno"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        summons = match.player_tables[self.position.player_idx].summons
        summon_name = "Fiery Sanctum Field"
        summon_exists = False
        for summon in summons:
            if summon.name == summon_name:
                summon_exists = True
                break
        if summon_exists:
            # make damage and create summon
            return super().get_actions(
                match,
                [
                    self.create_summon(summon_name),
                ],
            )
        return [self.create_summon(summon_name), self.charge_self(1)]


class LeonineBite(ElementalBurstBase):
    name: Literal["Leonine Bite"] = "Leonine Bite"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(
            match,
            [
                self.create_character_status("Incineration Drive"),
            ],
        )


class IncinerationDrive(ElementalBurstBase):
    name: Literal["Incineration Drive"] = "Incineration Drive"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        Always invalid for prepare skills
        """
        return False


# Talents


class StalwartAndTrue_4_1(SkillTalent):
    name: Literal["Stalwart and True"]
    version: Literal["4.1"] = "4.1"
    character_name: Literal["Dehya"] = "Dehya"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO,
        elemental_dice_number=4,
    )
    skill: Literal["Molten Inferno"] = "Molten Inferno"

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        If equipped, and dehya hp <= 6, heal 2 hp
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp <= 6:
            return [
                MakeDamageAction(
                    damage_value_list=[
                        DamageValue(
                            position=self.position,
                            damage_type=DamageType.HEAL,
                            target_position=character.position,
                            damage=-2,
                            damage_elemental_type=DamageElementalType.HEAL,
                            cost=self.cost.copy(),
                        )
                    ],
                )
            ]
        return []


# character base


class Dehya_4_1(CharacterBase):
    name: Literal["Dehya"]
    version: Literal["4.1"] = "4.1"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | MoltenInferno | LeonineBite | IncinerationDrive
    ] = []
    faction: List[FactionType] = [FactionType.SUMERU, FactionType.THE_EREMITES]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Sandstorm Assault",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            MoltenInferno(),
            LeonineBite(),
            IncinerationDrive(),
        ]


register_class(Dehya_4_1 | StalwartAndTrue_4_1 | FierySanctumField_4_1)



import os
from tests.utils_for_test import do_log_tests


def test_balance():
    json_fname = "balance_thuncer_vourukashas_stove.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path, hp_modify=10)


def test_cryo_hypostasis():
    json_fname = "cryo_hypostasis.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def millennial_999_test(match, cmd):
    """
    p0c0 millennial is talent status 0
    """
    status = match.player_tables[0].characters[0].status[0]
    assert status.desc == "talent"


def test_millennial_pearl_seahorse():
    json_fname = "millennial_pearl_seahorse.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path, other_tests={999: millennial_999_test})


def test_millennial_pearl_seahorse_2():
    json_fname = "millennial_pearl_seahorse2.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_thoma():
    json_fname = "thoma.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_sayu():
    json_fname = "sayu.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_sapwood_machine_veteran():
    json_fname = "sapwood_machine_veteran.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_machine_2():
    json_fname = "machine_2.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_veteran_2():
    json_fname = "veteran_2.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_silver():
    json_fname = "silver.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_jeht_sunyata():
    json_fname = "jeht_sunyata.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


if __name__ == "__main__":
    # test_balance()
    # test_cryo_hypostasis()
    # test_millennial_pearl_seahorse()
    # test_millennial_pearl_seahorse_2()
    # test_thoma()
    # test_sayu()
    # test_sapwood_machine_veteran()
    # test_machine_2()
    # test_veteran_2()
    test_silver()



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import ObjectPositionType, SkillType, WeaponType

from ....action import ChargeAction
from ....event import SkillEndEventArguments
from ....struct import Cost
from .base import RoundEffectWeaponBase


class FavoniusBase(RoundEffectWeaponBase):
    name: str
    cost: Cost = Cost(same_dice_number=3)
    version: str
    weapon_type: WeaponType
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChargeAction]:
        """
        if self character use elemental skill, charge one more
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self character or not equipped
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        if self.usage <= 0:
            # no usage
            return []
        self.usage -= 1
        return [
            ChargeAction(
                player_idx=self.position.player_idx,
                character_idx=self.position.character_idx,
                charge=1,
            )
        ]


class FavoniusSword_3_7(FavoniusBase):
    name: Literal["Favonius Sword"]
    version: Literal["3.7"] = "3.7"
    weapon_type: Literal[WeaponType.SWORD] = WeaponType.SWORD


register_class(FavoniusSword_3_7)



from .server.interaction import (  # noqa: F401
    Responses,
    SwitchCardResponse,
    ChooseCharacterResponse,
    RerollDiceResponse,
    SwitchCharacterResponse,
    ElementalTuningResponse,
    DeclareRoundEndResponse,
    UseSkillResponse,
    UseCardResponse,
)



import logging
from typing import Tuple, List, Literal

from .agent_base import AgentBase
from .random_agent import RandomAgent
from ..server.match import Match
from ..server.interaction import (
    Requests,
    Responses,
    SwitchCardRequest,
    SwitchCardResponse,
    ChooseCharacterRequest,
    ChooseCharacterResponse,
    RerollDiceRequest,
    RerollDiceResponse,
    DeclareRoundEndRequest,
    DeclareRoundEndResponse,
    ElementalTuningRequest,
    ElementalTuningResponse,
    SwitchCharacterRequest,
    SwitchCharacterResponse,
    UseSkillRequest,
    UseSkillResponse,
    UseCardRequest,
    UseCardResponse,
)
from ..server.consts import DieColor
from ..server.dice import Dice


class InteractionAgent_V1_0(AgentBase):
    """
    Agent that read inputs interactively, and generate responses.
    All idx are start with 0.

    NOT THE NEWSET VERSION, ONLY FOR COMPATIBILITY.

    Command format and sample:
        sw_card: used to switch cards, with variable length of card indices.
            sample: sw_card 0 1 2
        choose: used to choose character, with one character idx.
            sample: choose 0
        reroll: used to reroll dice, with variable length of dice indices.
            sample: reroll 0 1 2
        end: used to declare round end, with no args.
            sample: end
        tune: used to elemental tuning, with one dice color and one card idx.
            sample: tune pyro 0
        sw_char: used to switch character, with one character idx and variable
                length of cost colors.
            sample: sw_char 0 pyro
        skill: used to use skill, with one skill idx and variable length of
                cost colors. skill idx is the index of the skill in the
                available requests.
            sample: skill 0 omni hydro geo
        card: used to use card, with one card idx, one target idx and variable
                length of cost colors. card idx is the index of the card in the
                available requests. target idx is the index of the target in
                the targets of requests, if no targets is needed, target idx
                can be any number.
            sample: card 0 0 omni geo

    Args:
        player_idx: The player idx of the agent.
        version: The version of the agent.
        _cmd_to_name: A dict that maps command to request name.
        verbose_level: The verbose level of the agent, higher means more
            verbose, 0 means no verbose. in level 1, when generate response,
            request names are shown. in level 2, when generate response,
            json of requests are shown.
        available_reqs: The available requests of the agent. Will be updated
            when generate response.
        commands: The commands of the agent. If it is not empty, when get
            command, it will be used instead of input.
        random_after_no_command: If True, when no command, generate random
            response by RandomAgent.
    """

    version: Literal["1.0"]
    _cmd_to_name = {
        "sw_card": "SwitchCardRequest",
        "choose": "ChooseCharacterRequest",
        "reroll": "RerollDiceRequest",
        "end": "DeclareRoundEndRequest",
        "tune": "ElementalTuningRequest",
        "sw_char": "SwitchCharacterRequest",
        "skill": "UseSkillRequest",
        "card": "UseCardRequest",
    }
    verbose_level: int = 0
    available_reqs: List[Requests] = []

    commands: List[str] = []
    only_use_command: bool = False

    random_after_no_command: bool = False
    random_agent: RandomAgent = RandomAgent(player_idx=-1)

    def _print_requests(self):
        if self.verbose_level == 2:
            print("available requests:")
            for i, req in enumerate(self.available_reqs):
                print(f"{i}: {req.name} {req.json()}")
            return
        print(f"available requests: {[x.name for x in self.available_reqs]}")

    def _get_input(self) -> str:
        if len(self.commands) > 0:
            cmd = self.commands[0]
            self.commands = self.commands[1:]
            logging.info(f"InteractionAgent {self.player_idx}: {cmd}")
            return cmd
        if self.only_use_command:
            return "no_command"
        print(f"InteractionAgent {self.player_idx}: ")
        return input()

    def _get_cmd(self) -> Tuple[str, List[str]]:
        while True:
            i = self._get_input().strip().split(" ")
            if len(i) == 0:
                continue
            cmd = i[0]
            if cmd == "help":
                print(f"available commands: {list(self._cmd_to_name.keys())}")
            elif cmd == "no_command":
                return "no_command", []
            elif cmd == "print":
                self._print_requests()
            elif cmd == "verbose":
                if len(i) == 1:
                    print(f"current verbose level: {self.verbose_level}")
                else:
                    self.verbose_level = int(i[1])
                    print(f"set verbose level to {self.verbose_level}")
            elif cmd in self._cmd_to_name:
                return self._cmd_to_name[cmd], i[1:]
            else:
                try:
                    cmd_num = int(cmd)
                    if cmd_num < 0 or cmd_num >= len(self.available_reqs):
                        print("invalid command number")
                        if self.only_use_command:
                            raise AssertionError()
                        continue
                    return self.available_reqs[cmd_num].name, i[1:]
                except ValueError:
                    print("invalid command")
                    if self.only_use_command:
                        raise AssertionError()
                    continue

    def _colors_to_idx(self, color_names: List[str], all: List[DieColor]) -> List[int]:
        res: List[int] = []
        selected = [DieColor[x.upper()] for x in color_names]
        all_e: List[DieColor | None] = list(all)
        for x in selected:
            res.append(all_e.index(x))
            all_e[all_e.index(x)] = None
        return res

    def generate_response(self, match: Match) -> Responses | None:
        if self.random_after_no_command and len(self.commands) == 0:
            self.random_agent.player_idx = self.player_idx
            return self.random_agent.generate_response(match)
        self.available_reqs = [
            x for x in match.requests if x.player_idx == self.player_idx
        ]
        if len(self.available_reqs) == 0:
            return None
        if self.verbose_level >= 1:
            self._print_requests()
        try:
            req_name, args = self._get_cmd()
            if req_name == "no_command":
                return None
            reqs = [x for x in self.available_reqs if x.name == req_name]
            if len(reqs) == 0:
                print(f"no such request: {req_name}")
                if self.only_use_command:
                    raise AssertionError()
                return self.generate_response(match)
            if req_name == "SwitchCardRequest":
                return self.resp_switch_card(args, reqs)  # type: ignore
            if req_name == "ChooseCharacterRequest":
                return self.resp_choose_character(args, reqs)  # type: ignore
            if req_name == "RerollDiceRequest":
                return self.resp_reroll_dice(args, reqs)  # type: ignore
            if req_name == "DeclareRoundEndRequest":
                return self.resp_declare_round_end(args, reqs)  # type: ignore
            if req_name == "ElementalTuningRequest":
                return self.resp_elemental_tuning(args, reqs)  # type: ignore
            if req_name == "SwitchCharacterRequest":
                return self.resp_switch_character(args, reqs)  # type: ignore
            if req_name == "UseSkillRequest":
                return self.resp_use_skill(args, reqs)  # type: ignore
            if req_name == "UseCardRequest":
                return self.resp_use_card(args, reqs)  # type: ignore
        except Exception as e:
            print(f"error: {e}")
            if self.only_use_command:
                raise e
            return self.generate_response(match)

    def resp_switch_card(
        self, args: List[str], reqs: List[SwitchCardRequest]
    ) -> SwitchCardResponse:
        """
        args: variable length of card indices.
        """
        assert len(reqs) == 1
        return SwitchCardResponse(request=reqs[0], card_idxs=[int(x) for x in args])

    def resp_choose_character(
        self, args: List[str], reqs: List[ChooseCharacterRequest]
    ) -> ChooseCharacterResponse:
        """
        args: one character idx.
        """
        assert len(reqs) == 1
        return ChooseCharacterResponse(request=reqs[0], character_idx=int(args[0]))

    def resp_reroll_dice(
        self, args: List[str], reqs: List[RerollDiceRequest]
    ) -> RerollDiceResponse:
        """
        args: variable length of dice indices.
        """
        assert len(reqs) == 1
        return RerollDiceResponse(
            request=reqs[0], reroll_dice_idxs=[int(x) for x in args]
        )

    def resp_elemental_tuning(
        self, args: List[str], reqs: List[ElementalTuningRequest]
    ) -> ElementalTuningResponse:
        """
        args: one dice color and one card idx.
        """
        assert len(reqs) == 1
        dice_idx = self._colors_to_idx(args[:1], reqs[0].dice_colors)[0]
        return ElementalTuningResponse(
            request=reqs[0], dice_idx=dice_idx, card_idx=int(args[1])
        )

    def resp_switch_character(
        self, args: List[str], reqs: List[SwitchCharacterRequest]
    ) -> SwitchCharacterResponse:
        """
        args: one character idx and variable length of cost colors.
        """
        cidx = int(args[0])
        selected_req: SwitchCharacterRequest | None = None
        for req in reqs:
            if req.target_character_idx == cidx:
                selected_req = req
                break
        assert selected_req is not None
        dice_idxs = self._colors_to_idx(args[1:], reqs[0].dice_colors)
        return SwitchCharacterResponse(request=selected_req, dice_idxs=dice_idxs)

    def resp_declare_round_end(
        self, args: List[str], reqs: List[DeclareRoundEndRequest]
    ) -> DeclareRoundEndResponse:
        """
        args: none
        """
        assert len(reqs) == 1
        return DeclareRoundEndResponse(request=reqs[0])

    def resp_use_skill(
        self, args: List[str], reqs: List[UseSkillRequest]
    ) -> UseSkillResponse:
        """
        args: one skill idx, variable length of cost colors.
        """
        skill_req = reqs[int(args[0])]
        dice_idxs = self._colors_to_idx(args[1:], skill_req.dice_colors)
        return UseSkillResponse(request=skill_req, dice_idxs=dice_idxs)

    def resp_use_card(
        self, args: List[str], reqs: List[UseCardRequest]
    ) -> UseCardResponse:
        """
        args: one card idx, one target idx, variable length of cost colors.
        """
        card_req = reqs[int(args[0])]
        target_idx = int(args[1])
        target = None
        if len(card_req.targets) > 0:
            target = card_req.targets[target_idx]
        dice_idxs = self._colors_to_idx(args[2:], card_req.dice_colors)
        return UseCardResponse(request=card_req, dice_idxs=dice_idxs, target=target)


class InteractionAgent_V2_0(InteractionAgent_V1_0):
    """
    This version changes the command format of tune, card and skill to make
    easier use. And can use idx of dice to instead die color.
    All idx are start with 0.

    For tune, it will need to input the card idx first, then the dice color.
    For skill, the skill idx is not the index of the skill in the available
        requests, but the index of the skill in the skills of the active
        character.
    For card, the card idx is not the index of the card in the available
        requests, but the index of the card in the hand.

    Changed commands:
        tune: used to elemental tuning, with one card idx and one dice color.
            sample: tune 0 pyro | tune 0 6
        skill: used to use skill, with one skill idx and variable length of
                cost colors. skill idx is the index of the skill of the active
                character.
            sample: skill 0 omni hydro geo | skill 2 0 2 3
        card: used to use card, with one card idx, one target idx and variable
                length of cost colors. card idx is the index of the card in
                hands. target idx is the index of the target in
                the targets of requests, if no targets is needed, target idx
                can be any number.
            sample: card 0 0 omni geo | card 1 0 5 6

    Unchanged commands (but can use idx of dice):
        sw_card: used to switch cards, with variable length of card indices.
            sample: sw_card 0 1 2
        choose: used to choose character, with one character idx.
            sample: choose 0
        reroll: used to reroll dice, with variable length of dice indices or
            dice colors.
            sample: reroll 0 1 2 | reroll pyro geo
        end: used to declare round end, with no args.
            sample: end
        sw_char: used to switch character, with one character idx and variable
                length of cost colors.
            sample: sw_char 0 pyro | sw_char 1 6

    Args:
        Unchanged, please refer to V1.0
    """

    version: Literal["2.0"] = "2.0"

    def _colors_to_idx(self, color_names: List[str], all: List[DieColor]) -> List[int]:
        try:
            color_idxs = [int(x) for x in color_names]
            # int representation
            return color_idxs
        except ValueError:
            pass
        # color representation
        dice = Dice(colors=all)
        res = dice.colors_to_idx([DieColor(x.upper()) for x in color_names])
        return res

    def resp_reroll_dice(
        self, args: List[str], reqs: List[RerollDiceRequest]
    ) -> RerollDiceResponse:
        """
        args: variable length of dice idx or colors
        """
        assert len(reqs) == 1
        dice_idx = self._colors_to_idx(args, reqs[0].colors)
        return RerollDiceResponse(request=reqs[0], reroll_dice_idxs=dice_idx)

    def resp_elemental_tuning(
        self, args: List[str], reqs: List[ElementalTuningRequest]
    ) -> ElementalTuningResponse:
        """
        args: one card idx and dice color.
        """
        assert len(reqs) == 1
        cost_idx = self._colors_to_idx(args[1:], reqs[0].dice_colors)[0]
        return ElementalTuningResponse(
            request=reqs[0], dice_idx=cost_idx, card_idx=int(args[0])
        )

    def resp_use_skill(
        self, args: List[str], reqs: List[UseSkillRequest]
    ) -> UseSkillResponse:
        """
        args: one skill idx, variable length of cost colors.
        """
        skill_reqs = [x for x in reqs if x.skill_idx == int(args[0])]
        assert len(skill_reqs) == 1
        skill_req = skill_reqs[0]
        dice_idxs = self._colors_to_idx(args[1:], skill_req.dice_colors)
        return UseSkillResponse(request=skill_req, dice_idxs=dice_idxs)

    def resp_use_card(
        self, args: List[str], reqs: List[UseCardRequest]
    ) -> UseCardResponse:
        """
        args: one card idx, one target idx, variable length of cost colors.
        """
        card_reqs = [x for x in reqs if x.card_idx == int(args[0])]
        assert len(card_reqs) == 1
        card_req = card_reqs[0]
        target_idx = int(args[1])
        target = None
        if len(card_req.targets) > 0:
            target = card_req.targets[target_idx]
        dice_idxs = self._colors_to_idx(args[2:], card_req.dice_colors)
        return UseCardResponse(request=card_req, dice_idxs=dice_idxs, target=target)


InteractionAgent = InteractionAgent_V2_0



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import DamageValue

from ....consts import DamageElementalType, DamageType, ObjectPositionType, SkillType

from ....action import MakeDamageAction

from ....event import SkillEndEventArguments

from ....struct import Cost
from .base import RoundEffectArtifactBase


class HealBySkillArtifactBase(RoundEffectArtifactBase):
    name: str
    version: str
    cost: Cost
    max_usage_per_round: int
    heal: int
    heal_skill_type: SkillType
    heal_target: Literal["SELF", "TEAM"]

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self use elemental skill, create one corresponding dice
        """
        if not (
            self.position.area == ObjectPositionType.CHARACTER
            and event.action.position.player_idx == self.position.player_idx
            and event.action.position.character_idx == self.position.character_idx
            and event.action.skill_type == self.heal_skill_type
            and self.usage > 0
        ):
            # not equipped or not self use elemental skill or no usage
            return []
        self.usage -= 1
        action = MakeDamageAction(
            damage_value_list=[],
        )
        if self.heal_target == "SELF":
            character = match.player_tables[self.position.player_idx].characters[
                self.position.character_idx
            ]
            action.damage_value_list.append(
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.HEAL,
                    target_position=character.position,
                    damage=-self.heal,
                    damage_elemental_type=DamageElementalType.HEAL,
                    cost=self.cost.copy(),
                )
            )
        else:
            characters = match.player_tables[self.position.player_idx].characters
            for character in characters:
                if character.is_defeated:
                    continue
                action.damage_value_list.append(
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-self.heal,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                )
        return [action]


class AdventurersBandana_3_3(HealBySkillArtifactBase):
    name: Literal["Adventurer's Bandana"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)
    max_usage_per_round: int = 3
    heal_skill_type: SkillType = SkillType.NORMAL_ATTACK
    heal: int = 1
    heal_target: Literal["SELF", "TEAM"] = "SELF"


class LuckyDogsSilverCirclet_3_3(HealBySkillArtifactBase):
    name: Literal["Lucky Dog's Silver Circlet"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(any_dice_number=2)
    max_usage_per_round: int = 1
    heal_skill_type: SkillType = SkillType.ELEMENTAL_SKILL
    heal: int = 2
    heal_target: Literal["SELF", "TEAM"] = "SELF"


class TravelingDoctorsHandkerchief_3_3(HealBySkillArtifactBase):
    name: Literal["Traveling Doctor's Handkerchief"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)
    max_usage_per_round: int = 1
    heal_skill_type: SkillType = SkillType.ELEMENTAL_BURST
    heal: int = 1
    heal_target: Literal["SELF", "TEAM"] = "TEAM"


register_class(
    AdventurersBandana_3_3
    | LuckyDogsSilverCirclet_3_3
    | TravelingDoctorsHandkerchief_3_3
)



from typing import Any, Dict, List, Literal

from ....action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    DrawCardAction,
)
from ....struct import ObjectPosition
from .....utils.class_registry import register_class
from ....summon.base import AttackerSummonBase
from ....modifiable_values import DamageDecreaseValue
from ....match import Match
from ....event import (
    DeclareRoundEndEventArguments,
    GameStartEventArguments,
    RoundPrepareEventArguments,
)
from ....status.character_status.base import DefendCharacterStatus
from ....struct import Cost
from ....consts import (
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    TalentBase,
)
from .....utils.desc_registry import DescDictType


class FontemerPearl_4_4(DefendCharacterStatus):
    name: Literal["Fontemer Pearl"] = "Fontemer Pearl"
    desc: Literal["", "talent"] = ""
    version: Literal["4.4"] = "4.4"
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1

    extra_usage: int = 1
    round_extra_usage: int = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # if marked as talent, increase extra usage.
        if self.desc == "talent":
            self.extra_usage = 2
            self.round_extra_usage = 2

    def renew(self, obj: "FontemerPearl_4_4") -> None:
        # When renew, update extra usage. If has talent, update desc
        self.usage = max(self.usage, obj.usage)
        self.extra_usage = max(self.extra_usage, obj.extra_usage)
        self.round_extra_usage = max(self.round_extra_usage, obj.round_extra_usage)
        if obj.desc == "talent":
            self.desc = "talent"

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.extra_usage = self.round_extra_usage
        return []

    def value_modifier_DAMAGE_DECREASE(
        self, value: DamageDecreaseValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> DamageDecreaseValue:
        initial_usage = self.usage
        ret = super().value_modifier_DAMAGE_DECREASE(value, match, mode)
        if (
            initial_usage > self.usage
            and value.position.area == ObjectPositionType.SUMMON
            and self.extra_usage > 0
        ):
            # When damage from summon and usage decreased and has extra usage
            self.usage += 1
            self.extra_usage -= 1
        return ret

    def event_handler_DECLARE_ROUND_END(
        self, event: DeclareRoundEndEventArguments, match: Any
    ) -> List[DrawCardAction]:
        """
        if declare round end, draw card
        """
        if (
            match.player_tables[self.position.player_idx].active_character_idx
            == self.position.character_idx
            and event.action.player_idx == self.position.player_idx
        ):
            # active character
            return [
                DrawCardAction(
                    player_idx=self.position.player_idx,
                    number=1,
                    draw_if_filtered_not_enough=True,
                )
            ]
        return []


class ResonantCoralOrb_4_4(AttackerSummonBase):
    name: Literal["Resonant Coral Orb"] = "Resonant Coral Orb"
    version: Literal["4.4"] = "4.4"
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    usage: int = 2
    max_usage: int = 2


class SwirlingSchoolOfFish(ElementalSkillBase):
    name: Literal["Swirling School of Fish"] = "Swirling School of Fish"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        ret = super().get_actions(match)
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        target_status = None
        for s in status:
            if s.name == "Fontemer Pearl":
                target_status = s
                break
        # if no status, only attack
        if target_status is None:
            return ret
        return ret + [
            ChangeObjectUsageAction(
                object_position=target_status.position, change_usage=1
            )
        ]


class FontemerHoarthunder(ElementalBurstBase):
    name: Literal["Fontemer Hoarthunder"] = "Fontemer Hoarthunder"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        args: Dict[str, Any] = {"usage": 2}
        if self.is_talent_equipped(match):
            args["desc"] = "talent"
        return super().get_actions(
            match,
            [
                self.create_character_status("Fontemer Pearl", args),
                self.create_summon("Resonant Coral Orb"),
            ],
        )


class PearlArmor(CreateStatusPassiveSkill):
    name: Literal["Pearl Armor"] = "Pearl Armor"
    status_name: Literal["Fontemer Pearl"] = "Fontemer Pearl"
    regenerate_when_revive: bool = False

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        When game begin, gain status
        """
        return [self.create_character_status(self.status_name, {"usage": 2})]


class PearlSolidification_4_4(TalentBase):
    name: Literal["Pearl Solidification"]
    version: Literal["4.4"] = "4.4"
    character_name: Literal["Millennial Pearl Seahorse"] = "Millennial Pearl Seahorse"
    cost: Cost = Cost()
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EQUIPMENT.value
    )
    remove_when_used: bool = False

    def get_actions(self, target: ObjectPosition | None, match: Match) -> List[Actions]:
        assert target is not None
        ret = super().get_actions(target, match)
        # create status again, so it will be updated to talent version
        ret += [
            CreateObjectAction(
                object_name="Fontemer Pearl",
                object_position=target.set_area(ObjectPositionType.CHARACTER_STATUS),
                object_arguments={"desc": "talent"},
            )
        ]
        status = (
            match.player_tables[target.player_idx]
            .characters[target.character_idx]
            .status
        )
        pearl_status = None
        for s in status:
            if s.name == "Fontemer Pearl":
                pearl_status = s
                break
        if pearl_status is not None:
            # status already existed, increase usage
            ret += [
                ChangeObjectUsageAction(
                    object_position=pearl_status.position, change_usage=1
                )
            ]
        return ret


class MillennialPearlSeahorse_4_4(CharacterBase):
    name: Literal["Millennial Pearl Seahorse"]
    version: Literal["4.4"] = "4.4"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 8
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | SwirlingSchoolOfFish
        | FontemerHoarthunder
        | PearlArmor
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Tail Sweep",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SwirlingSchoolOfFish(),
            FontemerHoarthunder(),
            PearlArmor(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Millennial Pearl Seahorse": {
        "names": {"en-US": "Millennial Pearl Seahorse", "zh-CN": "千年珍珠骏麟"},
        "descs": {"4.4": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_SeaHorsePrimo.png",  # noqa: E501
        "id": 2403,
    },
    "SKILL_Millennial Pearl Seahorse_NORMAL_ATTACK/Tail Sweep": {
        "names": {"en-US": "Tail Sweep", "zh-CN": "旋尾扇击"},
        "descs": {
            "4.4": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Millennial Pearl Seahorse_ELEMENTAL_SKILL/Swirling School of Fish": {  # noqa: E501
        "names": {"en-US": "Swirling School of Fish", "zh-CN": "霰舞鱼群"},
        "descs": {
            "4.4": {
                "en-US": "Deals 3 Electro DMG. Once per Round: If this character has Fontemer Pearl attached, that Usage(s) +1.",  # noqa: E501
                "zh-CN": "造成3点雷元素伤害。\n每回合1次：如果本角色已附属原海明珠，则使其可用次数+1。",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Fontemer Pearl": {
        "names": {"en-US": "Fontemer Pearl", "zh-CN": "原海明珠"},
        "descs": {
            "4.4": {
                "en-US": "When the character to which this is attached takes DMG: Negate 1 DMG. Once per Round, Usage(s) will not be used when negating DMG from Summons.\nUsage(s): 2\nWhen you declare the end of your Round: If the character to which this is attached is the active character, draw 1 card.",  # noqa: E501
                "zh-CN": "所附属角色受到伤害时：抵消1点伤害；每回合1次，抵消来自召唤物的伤害时不消耗可用次数。\n可用次数：2\n\n我方宣布结束时：如果所附属角色为「出战角色」，则抓1张牌。",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Fontemer Pearl_talent": {
        "names": {"en-US": "Fontemer Pearl", "zh-CN": "原海明珠"},
        "descs": {
            "4.4": {
                "en-US": "When the character to which this is attached takes DMG: Negate 1 DMG. Twice per Round, Usage(s) will not be used when negating DMG from Summons.\nUsage(s): 2\nWhen you declare the end of your Round: If the character to which this is attached is the active character, draw 1 card.",  # noqa: E501
                "zh-CN": "所附属角色受到伤害时：抵消1点伤害；每回合2次，抵消来自召唤物的伤害时不消耗可用次数。\n可用次数：2\n\n我方宣布结束时：如果所附属角色为「出战角色」，则抓1张牌。",  # noqa: E501
            }
        },
    },
    "SKILL_Millennial Pearl Seahorse_ELEMENTAL_BURST/Fontemer Hoarthunder": {
        "names": {"en-US": "Fontemer Hoarthunder", "zh-CN": "原海古雷"},
        "descs": {
            "4.4": {
                "en-US": "Deals 1 Electro DMG, attaches Fontemer Pearl to this character, and summons 1 Resonant Coral Orb.",  # noqa: E501
                "zh-CN": "造成1点雷元素伤害，本角色附属原海明珠，召唤共鸣珊瑚珠。",  # noqa: E501
            }
        },
    },
    "SUMMON/Resonant Coral Orb": {
        "names": {"en-US": "Resonant Coral Orb", "zh-CN": "共鸣珊瑚珠"},
        "descs": {
            "4.4": {
                "en-US": "End Phase: Deal 1 Electro DMG.\nUsage(s): 2",
                "zh-CN": "结束阶段：造成1点雷元素伤害。\n可用次数：2",
            }
        },
        "image_path": "cardface/Summon_SeaHorsePrimo.png",
    },
    "SKILL_Millennial Pearl Seahorse_PASSIVE/Pearl Armor": {
        "names": {"en-US": "Pearl Armor", "zh-CN": "明珠甲胄"},
        "descs": {
            "4.4": {
                "en-US": "(Passive) When the battle begins, attach Fontemer Pearl to this character.",  # noqa: E501
                "zh-CN": "【被动】战斗开始时，本角色附属原海明珠。",  # noqa: E501
            }
        },
    },
    "TALENT_Millennial Pearl Seahorse/Pearl Solidification": {
        "names": {"en-US": "Pearl Solidification", "zh-CN": "明珠固化"},
        "descs": {
            "4.4": {
                "en-US": "Can only be played if your active character is Millennial Pearl Seahorse: When entering play, Millennial Pearl Seahorse will have Fontemer Pearl with 1 Usage(s) attached to them. If they already have Fontemer Pearl attached, Usage(s) +1 instead.\nWhen the Fontemer Pearl attached to Millennial Pearl Seahorse negates DMG from Summons, change to Usage(s) not being consumed twice per Round.\n(Your deck must contain Millennial Pearl Seahorse to add this card to your deck)",  # noqa: E501
                "zh-CN": "我方出战角色为千年珍珠骏麟时，才能打出：入场时，使千年珍珠骏麟附属可用次数为1的原海明珠；如果已附属原海明珠，则使其可用次数+1。\n装备有此牌的千年珍珠骏麟所附属的原海明珠抵消召唤物伤害时，改为每回合2次不消耗可用次数。\n（牌组中包含千年珍珠骏麟，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_SeaHorsePrimo.png",  # noqa: E501
        "id": 224031,
    },
}


register_class(
    FontemerPearl_4_4
    | ResonantCoralOrb_4_4
    | MillennialPearlSeahorse_4_4
    | PearlSolidification_4_4,
    desc,
)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....action import Actions, ChargeAction

from ....action import ChangeObjectUsageAction, CreateObjectAction
from ....event import (
    ChargeEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)

from ....modifiable_values import DamageIncreaseValue
from ....status.character_status.base import ShieldCharacterStatus
from ....status.team_status.base import ShieldTeamStatus

from ....consts import FactionType, ObjectPositionType, SkillType, WeaponType

from ....struct import Cost
from .base import RoundEffectWeaponBase, WeaponBase


class VortexVanquisher_3_7(RoundEffectWeaponBase):
    name: Literal["Vortex Vanquisher"]
    cost: Cost = Cost(same_dice_number=3)
    version: Literal["3.7"] = "3.7"
    weapon_type: WeaponType = WeaponType.POLEARM
    max_usage_per_round: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If self has shield, +1 damage.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not current character using skill
            return value
        # modify damage
        assert mode == "REAL"
        value.damage += self.damage_increase
        # if have shield, increase one more
        have_shield = False
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if issubclass(type(status), ShieldTeamStatus):
                have_shield = True
        if not have_shield:
            # no shield team status, check character status
            character = match.player_tables[self.position.player_idx].characters[
                self.position.character_idx
            ]
            for status in character.status:
                if issubclass(type(status), ShieldCharacterStatus):
                    have_shield = True
        if have_shield:
            value.damage += 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        if self character use elemental skill, check whether have
        shield team status. If so, add 1 usage.
        """
        if event.action.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        if self.usage <= 0:
            # no usage
            return []
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self character use skill or not equipped
            return []
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if issubclass(type(status), ShieldTeamStatus):
                # find shield status, add 3 usage
                self.usage -= 1
                return [
                    ChangeObjectUsageAction(
                        object_position=status.position,
                        change_usage=1,
                    )
                ]
        return []


class LithicSpear_3_7(WeaponBase):
    name: Literal["Lithic Spear"]
    cost: Cost = Cost(same_dice_number=3)
    version: Literal["3.7"] = "3.7"
    weapon_type: WeaponType = WeaponType.POLEARM

    def generate_shield(self, count: int) -> List[Actions]:
        count = min(count, 3)
        if count > 0:
            return [
                CreateObjectAction(
                    object_position=self.position.set_area(
                        ObjectPositionType.CHARACTER_STATUS
                    ),
                    object_name=self.name,
                    object_arguments={
                        "usage": count,
                        "max_usage": count,
                    },
                )
            ]
        return []

    def equip(self, match: Any) -> List[Actions]:
        """
        Generate shield
        """
        count = 0
        for c in match.player_tables[self.position.player_idx].characters:
            if FactionType.LIYUE in c.faction:
                count += 1
        return self.generate_shield(count)


class LithicSpear_3_3(LithicSpear_3_7):
    version: Literal["3.3"]

    def equip(self, match: Any) -> List[Actions]:
        """
        Generate shield
        """
        count = 0
        for c in match.player_tables[self.position.player_idx].characters:
            if c.is_alive and FactionType.LIYUE in c.faction:
                count += 1
        return self.generate_shield(count)


class EngulfingLightning_3_7(RoundEffectWeaponBase):
    name: Literal["Engulfing Lightning"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=3)
    weapon_type: WeaponType = WeaponType.POLEARM
    max_usage_per_round: int = 1

    def _charge_one(self, match: Any) -> List[ChargeAction]:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.charge == 0 and self.usage > 0:
            self.usage -= 1
            return [
                ChargeAction(
                    player_idx=self.position.player_idx,
                    character_idx=self.position.character_idx,
                    charge=1,
                )
            ]
        return []

    def equip(self, match: Any) -> List[Actions]:
        return super().equip(match) + self._charge_one(match)

    def event_handler_CHARGE(
        self, event: ChargeEventArguments, match: Any
    ) -> List[ChargeAction]:
        return self._charge_one(match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        ret = super().event_handler_ROUND_PREPARE(event, match)
        return ret + self._charge_one(match)


register_class(
    LithicSpear_3_7 | VortexVanquisher_3_7 | EngulfingLightning_3_7 | LithicSpear_3_3
)



import os
from tests.utils_for_test import do_log_tests


def test_eremite():
    json_fname = "test_eremite.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


if __name__ == "__main__":
    test_eremite()



import json
from lpsim.agents import InteractionAgent
from lpsim import Deck, Match, MatchState
from tests.utils_for_test import get_random_state, make_respond, remove_ids, set_16_omni


def get_memento_lens_match():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "sw_char 2 15",
            "card 1 0 14",
            "card 1 1 13 12",
            "sw_char 0 11",
            "end",
            "card 2 2",
            "card 2 2 15 14",
            "end",
            "card 1 0 15 14",
            "card 3 0 13 12",
            "card 1 0 11 10 9",
            "card 0 0 8",
            "card 0 0",
            "end",
            "sw_char 1 15",
        ],
        [
            "sw_card",
            "choose 1",
            "card 0 1 15 14",
            "sw_char 2 13",
            "card 1 2 12 11",
            "end",
            "card 3 0 15 14 13",
            "card 1 0 12",
            "sw_char 1 11",
            "end",
            "card 3 0 15",
            "card 2 0",
            "card 2 0 14",
            "card 0 0 13",
            "end",
            "card 0 0",
            "card 1 0 15 14 13",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        character:Ningguang@3.3
        character:Kaeya@3.3
        character:Chongyun@3.3
        Sacrificial Fragments@3.3*5
        Instructor's Cap@3.3*5
        Chinju Forest@3.7*5
        Paimon@3.3*5
        Memento Lens@4.3*5
        Abyssal Summons@3.3*5
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    return agent_0, agent_1, match


def test_memento_lens():
    agent_0, agent_1, match = get_memento_lens_match()
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_memento_lens_load_and_save():
    agent_0, agent_1, match = get_memento_lens_match()
    assert match.start()[0]
    match.step()
    histories = [match.json()]
    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        histories.append(match.json())
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break
    agent_0, agent_1, match = get_memento_lens_match()
    assert match.start()[0]
    match.step()
    for old_match_json in histories:  # pragma: no branch
        match_copy = match.copy(deep=True)
        assert remove_ids(match_copy) == remove_ids(
            Match(**json.loads(match_copy.json()))
        )
        assert remove_ids(match_copy) == remove_ids(Match(**json.loads(old_match_json)))
        match = Match(**json.loads(match.json()))
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_memento_lens()
    test_memento_lens_load_and_save()



from typing import Dict, List, Literal
from ....struct import Cost

from .....utils.class_registry import register_class
from ....consts import ELEMENT_TO_DIE_COLOR, DamageType, ObjectPositionType
from ....action import Actions, CreateDiceAction, DrawCardAction
from ....match import Match
from ....event import ReceiveDamageEventArguments
from ....card.equipment.artifact.base import RoundEffectArtifactBase
from .....utils.desc_registry import DescDictType


class VeteransVisage_4_4(RoundEffectArtifactBase):
    name: Literal["Veteran's Visage"]
    version: Literal["4.4"] = "4.4"
    cost: Cost = Cost(any_dice_number=2)
    max_usage_per_round: int = 2

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[Actions]:
        if self.usage <= 0:
            # already used
            return []
        if not self.position.check_position_valid(
            event.final_damage.target_position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            area_same=True,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not equipped, or not self receive damage
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp == 0:
            # character is dying
            return []
        if event.final_damage.damage_type not in [DamageType.DAMAGE, DamageType.HEAL]:
            # not damage or heal
            return []
        if self.usage == 2:
            # create die
            self.usage -= 1
            character = match.player_tables[self.position.player_idx].characters[
                self.position.character_idx
            ]
            return [
                CreateDiceAction(
                    player_idx=self.position.player_idx,
                    color=ELEMENT_TO_DIE_COLOR[character.element],
                    number=1,
                )
            ]
        else:
            # draw card
            self.usage -= 1
            return [
                DrawCardAction(
                    player_idx=self.position.player_idx,
                    number=1,
                    draw_if_filtered_not_enough=True,
                )
            ]


desc: Dict[str, DescDictType] = {
    "ARTIFACT/Veteran's Visage": {
        "names": {"en-US": "Veteran's Visage", "zh-CN": "老兵的容颜"},
        "descs": {
            "4.4": {
                "en-US": "After this character takes DMG or is healed: Based on the number of times this effect has been triggered this Round, the effects will be different.\nFirst Trigger: Create 1 Elemental Die of the character's Elemental Type.\nSecond Trigger: Draw 1 card.\n(A character can equip a maximum of 1 Artifact)",  # noqa: E501
                "zh-CN": "角色受到伤害或治疗后：根据本回合触发此效果的次数，执行不同的效果。\n第一次触发：生成1个此角色类型的元素骰。\n第二次触发：抓1张牌。\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Artifact_LaobinDanjian.png",  # noqa: E501
        "id": 312023,
    },
}


register_class(VeteransVisage_4_4, desc)



from typing import Dict, List, Literal
from pydantic import PrivateAttr

from ....action import Actions, CreateObjectAction
from ....modifiable_values import CostValue
from ....event import ReceiveDamageEventArguments
from ....status.character_status.base import CharacterStatusBase
from ....consts import CostLabels, DamageType, IconType, ObjectPositionType
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....object_base import EventCardBase
from .....utils.class_registry import register_class
from .....utils.desc_registry import DescDictType


class MachineAssemblyLineStatus_4_4(CharacterStatusBase):
    name: Literal["Machine Assembly Line"] = "Machine Assembly Line"
    version: Literal["4.4"] = "4.4"
    usage: int = 0
    max_usage: int = 2
    _card_cost_label: int = PrivateAttr(
        CostLabels.ARTIFACT.value | CostLabels.WEAPON.value
    )
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[Actions]:
        if not event.final_damage.is_corresponding_character_receive_damage(
            self.position, match, ignore_piercing=False
        ):
            # not self receive damage
            if not (
                event.final_damage.damage_type == DamageType.HEAL
                and self.position.check_position_valid(
                    event.final_damage.target_position,
                    match,
                    player_idx_same=True,
                    character_idx_same=True,
                )
            ):
                # and not heal self
                return []
        # add usage
        self.usage = min(self.usage + 1, self.max_usage)
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        When self equip weapon or artifact with cost fewer than usage
        """
        if not (
            value.position.player_idx == self.position.player_idx
            and value.cost.original_value is not None
            and value.cost.original_value.total_dice_cost <= self.usage
            and value.cost.label & self._card_cost_label != 0
        ):
            # not self, or original usage larger than self usage, or not weapon
            # or artifact
            return value
        total_cost = value.cost.total_dice_cost
        if total_cost == 0:
            # cost is 0, do nothing
            return value
        for _ in range(total_cost):
            value.cost.decrease_cost(None)
        if mode == "REAL":
            self.usage = 0
        return value


class MachineAssemblyLine_4_4(EventCardBase):
    name: Literal["Machine Assembly Line"]
    version: Literal["4.4"] = "4.4"
    cost: Cost = Cost()

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        # can quip on all self alive characters
        ret: List[ObjectPosition] = []
        for character in match.player_tables[self.position.player_idx].characters:
            if character.is_alive:
                ret.append(character.position)
        return ret

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[CreateObjectAction]:
        assert target is not None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=target.set_area(ObjectPositionType.CHARACTER_STATUS),
                object_arguments={},
            )
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER_STATUS/Machine Assembly Line": {
        "names": {"en-US": "Machine Assembly Line", "zh-CN": "机关铸成之链"},
        "descs": {
            "4.4": {
                "en-US": "Each time after this character is targeted by and takes DMG or is healed: Accumulate 1 point of Battle-Readiness (Max 2 points).\nWhen you play a Weapon or Artifact with an original cost of fewer Dice than you have Battle-Readiness points: Clear all Battle-Readiness points and play that card for free.",  # noqa: E501
                "zh-CN": "该角色每次受到伤害或治疗后：累积1点「备战度」（最多累积2点）。\n我方打出原本费用不多于「备战度」的「武器」或「圣遗物」时：移除所有「备战度」，以免费打出该牌。",  # noqa: E501
            }
        },
    },
    "CARD/Machine Assembly Line": {
        "names": {"en-US": "Machine Assembly Line", "zh-CN": "机关铸成之链"},
        "descs": {
            "4.4": {
                "en-US": "Each time after your character is targeted by and takes DMG or is healed: Accumulate 1 point of Battle-Readiness (Max 2 points).\nWhen you play a Weapon or Artifact with an original cost of fewer Dice than you have Battle-Readiness points: Clear all Battle-Readiness points and play that card for free.",  # noqa: E501
                "zh-CN": "目标我方角色每次受到伤害或治疗后：累积1点「备战度」（最多累积2点）。\n我方打出原本费用不多于「备战度」的「武器」或「圣遗物」时：移除所有「备战度」，以免费打出该牌。",  # noqa: E501
            }
        },
        "image_path": "cardface/Event_Event_Fatiaoji.png",  # noqa: E501
        "id": 332028,
    },
}


register_class(MachineAssemblyLineStatus_4_4 | MachineAssemblyLine_4_4, desc)



from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....action import CreateDiceAction, RemoveObjectAction
from ....match import Match
from ....event import PlayerActionStartEventArguments
from ....consts import ELEMENT_TO_DIE_COLOR, IconType, ObjectPositionType
from ....struct import Cost
from ....card.support.base import UsageWithRoundRestrictionSupportBase
from ....card.support.locations import LocationBase
from .....utils.desc_registry import DescDictType


class OperaEpiclese_4_3(LocationBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Opera Epiclese"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 3
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Match
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        If self action start, and has usage, and no elemental dice, create
        one omni element.
        """
        if not (
            self.position.area == ObjectPositionType.SUPPORT
            and event.player_idx == self.position.player_idx
            and self.has_usage()
        ):
            # not equipped, or not self, or no usage
            return []
        equip_costs = [0] * len(match.player_tables)
        for player_idx, player_table in enumerate(match.player_tables):
            for character in player_table.characters:
                equips = [character.weapon, character.artifact, character.talent]
                for equip in equips:
                    if equip is not None:
                        equip_costs[player_idx] += equip.cost.total_dice_cost
        if (
            equip_costs[self.position.player_idx]
            < equip_costs[1 - self.position.player_idx]
        ):
            # not less than opponent, return
            return []
        # create one element die of active character
        self.use()
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=ELEMENT_TO_DIE_COLOR[active_character.element],
            )
        ] + self.check_should_remove()


desc: Dict[str, DescDictType] = {
    "SUPPORT/Opera Epiclese": {
        "names": {"en-US": "Opera Epiclese", "zh-CN": "欧庇克莱歌剧院"},
        "descs": {
            "4.3": {
                "en-US": "Before you choose an action: If the original Elemental Dice cost of the cards equipped to your characters is not less than that of the opposing side, create 1 Elemental Die of your active character's element. (Once per Round)\nUsage(s): 3",  # noqa: E501
                "zh-CN": "我方选择行动前：如果我方角色所装备卡牌的原本元素骰费用总和不比对方更低，则生成1个出战角色类型的元素骰。（每回合1次）\n可用次数：3",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_Location_Gejuyuan.png",  # noqa: E501
        "id": 321017,
    },
}


register_class(OperaEpiclese_4_3, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import RoundPrepareEventArguments, SkillEndEventArguments

from ...action import Actions, ChargeAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class ClawAndThunder(ElementalSkillBase):
    name: Literal["Claw and Thunder"] = "Claw and Thunder"
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = ElementalSkillBase.get_cost(ElementType.ELECTRO)


class LightningFang(ElementalBurstBase):
    name: Literal["Lightning Fang"] = "Lightning Fang"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(
            match, [self.create_character_status("The Wolf Within")]
        )


# Talents
class Awakening_4_2(SkillTalent):
    name: Literal["Awakening"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Razor"] = "Razor"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )
    skill: Literal["Claw and Thunder"] = "Claw and Thunder"
    usage: int = 1
    max_usage: int = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        reset usage
        """
        self.usage = self.max_usage
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChargeAction]:
        """
        If equipped and use elemental skill, charge one of our electro
        character
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not equipped, or this character use skill
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        # find charge target
        table = match.player_tables[self.position.player_idx]
        characters = table.characters
        active = characters[table.active_character_idx]
        target_idx = -1
        if active.element == ElementType.ELECTRO and active.charge < active.max_charge:
            # active electro and not full charge
            target_idx = table.active_character_idx
        else:
            # check other characters
            for chraractor_idx, character in enumerate(characters):
                if (
                    character.element == ElementType.ELECTRO
                    and character.charge < character.max_charge
                    and character.is_alive
                ):
                    target_idx = chraractor_idx
                    break
        if target_idx == -1:
            # no target
            return []
        if self.usage <= 0:
            # no usage
            return []
        self.usage -= 1
        # charge target
        return [
            ChargeAction(
                player_idx=self.position.player_idx, character_idx=target_idx, charge=1
            )
        ]


# character base


class Razor_3_8(CharacterBase):
    name: Literal["Razor"]
    version: Literal["3.8"] = "3.8"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | ClawAndThunder | LightningFang] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Steel Fang",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ClawAndThunder(),
            LightningFang(),
        ]


register_class(Razor_3_8 | Awakening_4_2)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_ganyu():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 2 0 15 14 13 12 11",
            "skill 2 10 9 8 7 6",
            "TEST 1 5 5 5 5 6 5",
            "sw_char 1 5",
            "end",
            "card 3 0 15 14 13 12 11",
            "skill 2 10 9 8 7 6",
            "TEST 1 4 4 4 0 2 0",
            "end",
            "sw_char 0 15",
            "skill 1 14 13 12",
            "skill 3 11 10 9",
        ],
        [
            "sw_card 2 3 4",
            "choose 1",
            "TEST 1 10 10 10 8 8 8",
            "skill 2 15 14 13 12 11",
            "TEST 1 8 8 8 5 6 5",
            "card 2 0 10 9 8 7 6",
            "sw_char 0 5",
            "end",
            "TEST 1 5 5 5 3 4 3",
            "sw_char 1 15",
            "TEST 1 5 5 5 0 2 0",
            "skill 3 14 13 12",
            "end",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "TEST 1 2 3 3 0 1 0",
            "skill 2 9 8 7 6 5",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Ganyu
        character:Ganyu@3.3
        character:Fischl
        Undivided Heart*15
        Undivided Heart@3.3*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_ganyu_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 1 9 8 7",
            "sw_char 1 6",
            "end",
            "skill 0 15 14 13",
            "choose 0",
            "end",
            "TEST 1 6 0 2 10 5 10",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "skill 2 15 14 13 12 11",
            "card 2 0 10 9 8 7 6",
            "skill 3 5 4 3",
            "skill 0 2 1 0",
            "end",
            "skill 0 15 14 13",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Ganyu
        character:Ganyu@3.3
        character:Fischl
        Undivided Heart*15
        Undivided Heart@3.3*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_ganyu_with_heavy_and_cryo_resonance():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "card 1 0 15",
            "card 1 0 14",
            "skill 2 13 12 11 10 9",
            "TEST 1 10 7 10 8 4 8",
            "skill 0 8 7 6",
        ],
        [
            "sw_card 4 3",
            "choose 1",
            "TEST 1 10 10 10 8 4 8",
            "card 1 0 15",
            "card 2 0 14",
            "skill 1 13 12 11",
            "TEST 1 10 7 10 8 3 8",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Ganyu@4.2*3
        Heavy Strike*15
        Elemental Resonance: Shattering Ice*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_ganyu_mona():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "skill 2 6 5 4",
            "sw_char 0 3",
            "end",
            "skill 2 15 14 13 12 11",
            "TEST 1 10 7 10 8 1 6",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "sw_char 2 6",
            "end",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Ganyu
        character:Mona
        character:Nahida
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_ganyu()
    test_ganyu_with_heavy_and_cryo_resonance()
    test_ganyu_mona()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...consts import DamageElementalType
from .xingqiu_3_6 import Xingqiu_3_6 as X_3_6
from .xingqiu_3_6 import Raincutter as R_3_6
from .xingqiu_4_1 import GuhuaStyle, FatalRainscreen


class Raincutter(R_3_6):
    name: Literal["Raincutter"] = "Raincutter"

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, application and create version 3.3 object
        """
        return [
            self.charge_self(-2),
            self.attack_opposite_active(
                match,
                self.damage,
                self.damage_type,
                [self.create_team_status("Rainbow Bladework", {"version": "3.3"})],
            ),
            self.element_application_self(match, DamageElementalType.HYDRO),
        ]


class Xingqiu_3_3(X_3_6):
    version: Literal["3.3"]
    skills: List[GuhuaStyle | FatalRainscreen | Raincutter] = []

    def _init_skills(self) -> None:
        self.skills = [GuhuaStyle(), FatalRainscreen(), Raincutter()]


register_class(Xingqiu_3_3)



from typing import Dict, List, Literal

from ....action import (
    ActionTypes,
    Actions,
    CreateObjectAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from .....utils.class_registry import register_class
from ....event import SkillEndEventArguments
from ....status.team_status.base import (
    ExtraAttackTeamStatus,
    ShieldTeamStatus,
    UsageTeamStatus,
)
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class BlazingBarrier_4_4(ShieldTeamStatus):
    name: Literal["Blazing Barrier"] = "Blazing Barrier"
    version: Literal["4.4"] = "4.4"
    usage: int = 1
    max_usage: int = 3


class ScorchingOoyoroi_4_4(ExtraAttackTeamStatus, UsageTeamStatus):
    name: Literal["Scorching Ooyoroi"] = "Scorching Ooyoroi"
    version: Literal["4.4"] = "4.4"
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    trigger_skill_type: SkillType = SkillType.NORMAL_ATTACK
    usage: int = 2
    max_usage: int = 2
    decrease_usage: bool = True
    icon_type: IconType = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        skill = match.get_object(event.action.position)
        assert skill is not None
        if skill.name == "Crimson Ooyoroi":
            # Thoma use burst Crimson Ooyoroi
            return []
        ret: List[MakeDamageAction | RemoveObjectAction] = []
        ret += super().event_handler_SKILL_END(event, match)
        if len(ret) == 0:
            return ret
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        damage_action.create_objects.append(
            CreateObjectAction(
                object_name="Blazing Barrier",
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        )
        return ret


class BlazingBlessing(ElementalSkillBase):
    name: Literal["Blazing Blessing"] = "Blazing Blessing"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match, [self.create_team_status("Blazing Barrier")])


class CrimsonOoyoroi(ElementalBurstBase):
    name: Literal["Crimson Ooyoroi"] = "Crimson Ooyoroi"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        status_args = {
            "usage": 2,
            "max_usage": 2,
        }
        if self.is_talent_equipped(match):
            status_args = {
                "usage": 3,
                "max_usage": 3,
            }
        return super().get_actions(
            match,
            [
                self.create_team_status("Blazing Barrier"),
                self.create_team_status("Scorching Ooyoroi", status_args),
            ],
        )


class ASubordinatesSkills_4_4(SkillTalent):
    name: Literal["A Subordinate's Skills"]
    version: Literal["4.4"] = "4.4"
    character_name: Literal["Thoma"] = "Thoma"
    skill: Literal["Crimson Ooyoroi"] = "Crimson Ooyoroi"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )


class Thoma_4_4(CharacterBase):
    name: Literal["Thoma"]
    version: Literal["4.4"] = "4.4"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | BlazingBlessing | CrimsonOoyoroi] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Swiftshatter Spear",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            BlazingBlessing(),
            CrimsonOoyoroi(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Thoma": {
        "names": {"en-US": "Thoma", "zh-CN": "托马"},
        "descs": {"4.4": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Avatar_Tohma.png",  # noqa: E501
        "id": 1311,
    },
    "SKILL_Thoma_NORMAL_ATTACK/Swiftshatter Spear": {
        "names": {"en-US": "Swiftshatter Spear", "zh-CN": "迅破枪势"},
        "descs": {
            "4.4": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Thoma_ELEMENTAL_SKILL/Blazing Blessing": {
        "names": {"en-US": "Blazing Blessing", "zh-CN": "烈烧佑命之侍护"},
        "descs": {
            "4.4": {
                "en-US": "Deals 2 Pyro DMG, creates 1 Blazing Barrier.",
                "zh-CN": "造成2点火元素伤害，生成烈烧佑命护盾。",
            }
        },
    },
    "TEAM_STATUS/Blazing Barrier": {
        "names": {"en-US": "Blazing Barrier", "zh-CN": "烈烧佑命护盾"},
        "descs": {
            "4.4": {
                "en-US": "Shield\nGrants 1 Shield point to your active character. (Can stack. Max 3 points.)",  # noqa: E501
                "zh-CN": "护盾\n为我方出战角色提供1点护盾。（可叠加，最多叠加到3点）",  # noqa: E501
            }
        },
    },
    "SKILL_Thoma_ELEMENTAL_BURST/Crimson Ooyoroi": {
        "names": {"en-US": "Crimson Ooyoroi", "zh-CN": "真红炽火之大铠"},
        "descs": {
            "4.4": {
                "en-US": "Deals 2 Pyro DMG, creates 1 Blazing Barrier and Scorching Ooyoroi.",  # noqa: E501
                "zh-CN": "造成2点火元素伤害，生成烈烧佑命护盾和炽火大铠。",
            }
        },
    },
    "TEAM_STATUS/Scorching Ooyoroi": {
        "names": {"en-US": "Scorching Ooyoroi", "zh-CN": "炽火大铠"},
        "descs": {
            "4.4": {
                "en-US": "After your character uses a Normal Attack: Deal 1 Pyro DMG and create a Blazing Barrier.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "我方角色普通攻击后：造成1点火元素伤害，生成烈烧佑命护盾。\n可用次数：2",  # noqa: E501
            }
        },
        "image_path": "status/Tohma_E.png",
    },
    "TALENT_Thoma/A Subordinate's Skills": {
        "names": {"en-US": "A Subordinate's Skills", "zh-CN": "僚佐的才巧"},
        "descs": {
            "4.4": {
                "en-US": "Combat Action: When your active character is Thoma, equip this card.\nAfter Thoma equips this card, immediately use Crimson Ooyoroi once.\nWhen your Thoma, who has this card equipped, creates a Scorching Ooyoroi, its starting Usage(s) +1.\n(You must have Thoma in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为托马时，装备此牌。\n托马装备此牌后，立刻使用一次真红炽火之大铠。\n装备有此牌的托马生成的炽火大铠，初始可用次数+1。\n（牌组中包含托马，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Constellation_Thoma.png",  # noqa: E501
        "id": 213111,
    },
}


register_class(
    BlazingBarrier_4_4 | ScorchingOoyoroi_4_4 | ASubordinatesSkills_4_4 | Thoma_4_4,
    desc,
)



from typing import Dict, List, Literal

from ....status.team_status.base import RoundTeamStatus
from .....utils.class_registry import register_class
from ....action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    DrawCardAction,
    MakeDamageAction,
)
from ....event import (
    RoundEndEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from ....summon.base import AttackerSummonBase
from ....modifiable_values import DamageIncreaseValue
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class GeneralsWarBanner_4_3(RoundTeamStatus):
    name: Literal["General's War Banner"] = "General's War Banner"
    version: Literal["4.3"] = "4.3"
    usage: int = 2
    max_usage: int = 3

    icon_type: IconType = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if (
            value.position.area != ObjectPositionType.SKILL
            and value.position.area != ObjectPositionType.CHARACTER
        ):
            # not damage from character
            return value
        if value.position.player_idx != self.position.player_idx:
            # not damage from self
            return value
        if (
            value.damage_type != DamageType.DAMAGE
            or value.damage_elemental_type != DamageElementalType.GEO
        ):
            # not geo damage
            return value
        # increase damage
        assert mode == "REAL"
        value.damage += 1
        return value


class GeneralsGlory_4_3(AttackerSummonBase):
    name: Literal["General's Glory"] = "General's Glory"
    version: Literal["4.3"] = "4.3"
    usage: int = 2
    max_usage: int = 2
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.GEO

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[MakeDamageAction | ChangeObjectUsageAction | CreateObjectAction]:
        ret: List[MakeDamageAction | ChangeObjectUsageAction | CreateObjectAction] = []
        ret += super().event_handler_ROUND_END(event, match)
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        characters = match.player_tables[self.position.player_idx].characters
        geo_number = 0
        for character in characters:
            if character.element == ElementType.GEO:
                geo_number += 1
        if geo_number >= 2:
            damage_action.create_objects.append(
                CreateObjectAction(
                    object_name="Crystallize",
                    object_position=ObjectPosition(
                        player_idx=self.position.player_idx,
                        area=ObjectPositionType.TEAM_STATUS,
                        id=-1,
                    ),
                    object_arguments={},
                )
            )
        return ret


class InuzakaAllRoundDefense(ElementalSkillBase):
    name: Literal["Inuzaka All-Round Defense"] = "Inuzaka All-Round Defense"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO,
        elemental_dice_number=3,
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(
            match, [self.create_team_status("General's War Banner")]
        )


class JuugaForwardUntoVictory(ElementalBurstBase):
    name: Literal["Juuga: Forward Unto Victory"] = "Juuga: Forward Unto Victory"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(
            match,
            [
                self.create_team_status("General's War Banner"),
                self.create_summon("General's Glory"),
            ],
        )


class RushingHoundSwiftAsTheWind_4_3(SkillTalent):
    name: Literal["Rushing Hound: Swift as the Wind"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Gorou"] = "Gorou"
    skill: Literal["Inuzaka All-Round Defense"] = "Inuzaka All-Round Defense"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO,
        elemental_dice_number=3,
    )
    usage: int = 1
    draw_card: bool = False

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        """
        Reset draw_card_usage
        """
        self.usage = 1
        self.draw_card = False
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return value
        if (
            value.position.area != ObjectPositionType.SKILL
            and value.position.area != ObjectPositionType.CHARACTER
        ):
            # not damage from character
            return value
        if value.position.player_idx != self.position.player_idx:
            # not damage from self
            return value
        if (
            value.damage_type != DamageType.DAMAGE
            or value.damage_elemental_type != DamageElementalType.GEO
        ):
            # not geo damage
            return value
        if self.usage <= 0:
            # no usage
            return value
        assert mode == "REAL"
        self.draw_card = True
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[DrawCardAction]:
        """
        If draw_card is True, and has General's War Banner in play,
        and has usage, draw a card and reset draw_card to False.
        """
        if not self.draw_card:
            return []
        self.draw_card = False
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if status.name == "General's War Banner":
                break
        else:
            return []
        assert self.usage > 0
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ]


class Gorou_4_3(CharacterBase):
    name: Literal["Gorou"]
    version: Literal["4.3"]
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | InuzakaAllRoundDefense | JuugaForwardUntoVictory
    ] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Ripping Fang Fletching",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            InuzakaAllRoundDefense(),
            JuugaForwardUntoVictory(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Gorou": {
        "names": {"en-US": "Gorou", "zh-CN": "五郎"},
        "image_path": "cardface/Char_Avatar_Gorou.png",  # noqa: E501
        "id": 1606,
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
    },
    "SKILL_Gorou_NORMAL_ATTACK/Ripping Fang Fletching": {
        "names": {"en-US": "Ripping Fang Fletching", "zh-CN": "呲牙裂扇箭"},
        "descs": {
            "4.3": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Gorou_ELEMENTAL_SKILL/Inuzaka All-Round Defense": {
        "names": {"en-US": "Inuzaka All-Round Defense", "zh-CN": "犬坂吠吠方圆阵"},
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Geo DMG, creates 1 General's War Banner.",
                "zh-CN": "造成2点岩元素伤害，生成大将旗指物。",
            }
        },
    },
    "TEAM_STATUS/General's War Banner": {
        "names": {"en-US": "General's War Banner", "zh-CN": "大将旗指物"},
        "descs": {
            "4.3": {
                "en-US": "Your party's characters deal +1 Geo DMG.\nDuration (Rounds): 2 (Can stack, Can stack till 3 Rounds)",  # noqa: E501
                "zh-CN": "我方角色造成的岩元素伤害+1。\n持续回合：2（可叠加，最多叠加到3回合）",  # noqa: E501
            }
        },
        "image_path": "status/Gorou_E.png",
    },
    "SKILL_Gorou_ELEMENTAL_BURST/Juuga: Forward Unto Victory": {
        "names": {"en-US": "Juuga: Forward Unto Victory", "zh-CN": "兽牙逐突形胜战法"},
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Geo DMG, creates 1 General's War Banner, summons General's Glory.",  # noqa: E501
                "zh-CN": "造成2点岩元素伤害，生成大将旗指物，召唤大将威仪。",
            }
        },
    },
    "SUMMON/General's Glory": {
        "names": {"en-US": "General's Glory", "zh-CN": "大将威仪"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deal 1 Geo DMG. If your party has 2 Geo characters, create 1 Crystallize.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "结束阶段：造成1点岩元素伤害；如果队伍中存在2名岩元素角色，则生成结晶。\n可用次数：2",  # noqa: E501
            }
        },
    },
    "TALENT_Gorou/Rushing Hound: Swift as the Wind": {
        "names": {"en-US": "Rushing Hound: Swift as the Wind", "zh-CN": "犬奔·疾如风"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Gorou, equip this card.\nAfter Gorou equips this card, immediately use Inuzaka All-Round Defense once.\nWhen your Gorou, who has this card equipped, is on the field, and after any of your characters deals Geo DMG: If General's War Banner is in play, draw 1 card. (Once per Round)\n(You must have Gorou in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为五郎时，装备此牌。\n五郎装备此牌后，立刻使用一次犬坂吠吠方圆阵。\n装备有此牌的五郎在场时，我方角色造成岩元素伤害后：如果场上存在大将旗指物，抓1张牌。（每回合1次）\n（牌组中包含五郎，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Constellation_Gorou.png",  # noqa: E501
        "id": 216061,
    },
}


register_class(
    GeneralsGlory_4_3
    | GeneralsWarBanner_4_3
    | RushingHoundSwiftAsTheWind_4_3
    | Gorou_4_3,
    desc,
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import SummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import (
    Actions,
    ChangeObjectUsageAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Character status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Summons


class LightfallSword_3_8(SummonBase):
    name: Literal["Lightfall Sword"] = "Lightfall Sword"
    version: Literal["3.8"] = "3.8"
    usage: int = 0
    max_usage: int = 999
    damage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.PHYSICAL
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        When round end, make damage to the opponent and remove self
        """
        player_idx = self.position.player_idx
        target_table = match.player_tables[1 - player_idx]
        target_character = target_table.get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target_character.position,
                        damage=self.damage + self.usage,
                        damage_elemental_type=self.damage_elemental_type,
                        cost=Cost(),
                    )
                ],
            ),
            RemoveObjectAction(
                object_position=self.position,
            ),
        ]


# Skills


class FavoniusBladeworkEdel(PhysicalNormalAttackBase):
    name: Literal["Favonius Bladework - Edel"] = "Favonius Bladework - Edel"
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.CRYO)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If summon Lightfall Sword exists, not charge but add two usage.
        """
        summons = match.player_tables[self.position.player_idx].summons
        LS_idx = -1
        for sidx, summon in enumerate(summons):
            if summon.name == "Lightfall Sword":
                LS_idx = sidx
                break
        ret: List[Actions] = []
        ret.append(self.attack_opposite_active(match, self.damage, self.damage_type))
        if LS_idx != -1:
            # add LS usage
            ret.append(
                ChangeObjectUsageAction(
                    object_position=summons[LS_idx].position,
                    change_usage=2,
                )
            )
        else:
            # charge
            ret.append(self.charge_self(1))
        return ret


class IcetideVortex(ElementalSkillBase):
    name: Literal["Icetide Vortex"] = "Icetide Vortex"
    version: Literal["3.8"] = "3.8"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If summon Lightfall Sword exists, not charge but add two or three
        usage, based on whether has talent.
        """
        summons = match.player_tables[self.position.player_idx].summons
        LS_idx = -1
        for sidx, summon in enumerate(summons):
            if summon.name == "Lightfall Sword":
                LS_idx = sidx
                break
        ret: List[Actions] = []
        ret.append(self.attack_opposite_active(match, self.damage, self.damage_type))
        if LS_idx != -1:
            # add LS usage
            usage = 2
            if self.is_talent_equipped(match):
                # has talent, add one more
                usage = 3
            ret.append(
                ChangeObjectUsageAction(
                    object_position=summons[LS_idx].position,
                    change_usage=usage,
                )
            )
        else:
            # charge
            ret.append(self.charge_self(1))
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if s.name == "Grimheart":
                # has Grimheart, return
                return ret
        # no Grimheart, add Grimheart
        ret[0].create_objects = [
            self.create_character_status("Grimheart", {"version": self.version})
        ]
        return ret


class GlacialIllumination(ElementalBurstBase):
    name: Literal["Glacial Illumination"] = "Glacial Illumination"
    version: Literal["3.8"] = "3.8"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and summon
        """
        return super().get_actions(
            match, [self.create_summon("Lightfall Sword", {"version": self.version})]
        )


# Talents


class WellspringOfWarLust_3_5(SkillTalent):
    name: Literal["Wellspring of War-Lust"]
    version: Literal["3.5"] = "3.5"
    character_name: Literal["Eula"] = "Eula"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=2
    )
    skill: Literal["Glacial Illumination"] = "Glacial Illumination"


# character base


class Eula_3_8(CharacterBase):
    name: Literal["Eula"]
    version: Literal["3.8"] = "3.8"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[FavoniusBladeworkEdel | IcetideVortex | GlacialIllumination] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [FavoniusBladeworkEdel(), IcetideVortex(), GlacialIllumination()]


register_class(Eula_3_8 | WellspringOfWarLust_3_5 | LightfallSword_3_8)



from typing import Dict, List, Literal

from ....action import ActionTypes, Actions, MakeDamageAction, RemoveObjectAction
from .....utils.class_registry import register_class
from ....status.character_status.base import CharacterStatusBase
from ....summon.base import AttackerSummonBase
from ....modifiable_values import DamageIncreaseValue, DamageValue
from ....event import RoundPrepareEventArguments, SkillEndEventArguments
from ....match import Match
from ....struct import Cost
from ....consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    ObjectType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class PropSurplus_4_3(CharacterStatusBase):
    name: Literal["Prop Surplus"] = "Prop Surplus"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 3
    icon_type: IconType = IconType.BUFF

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Match,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill
            return value
        source: SkillBase = match.get_object(value.position)  # type: ignore
        assert source is not None and source.type == ObjectType.SKILL
        if source.name != "Bewildering Lights":
            # not Bewildering Lights
            return value
        # increase damage by usage
        assert self.usage > 0
        assert mode == "REAL"
        value.damage += self.usage
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If self use Bewildering Lights, heal self and remove self.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill
            return []
        source = match.get_object(event.action.position)
        assert source is not None and source.type == ObjectType.SKILL
        if source.name != "Bewildering Lights":
            # not Bewildering Lights
            return []
        # heal self and remove self
        assert self.usage > 0
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-self.usage,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ]
            ),
            RemoveObjectAction(
                object_position=self.position,
            ),
        ]


class GrinMalkinHat_4_3(AttackerSummonBase):
    name: Literal["Grin-Malkin Hat"] = "Grin-Malkin Hat"
    version: Literal["4.3"] = "4.3"
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    usage: int = 1
    max_usage: int = 2
    renew_type: Literal["ADD"] = "ADD"


class PropArrow(ElementalNormalAttackBase):
    name: Literal["Prop Arrow"] = "Prop Arrow"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        ret = super().get_actions(match)
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        damage_action.create_objects += [
            self.create_summon("Grin-Malkin Hat"),
            self.create_character_status("Prop Surplus"),
        ]
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp >= 6:
            ret.append(self.attack_self(match, 1))
        return ret


class WondrousTrickMiracleParade(ElementalBurstBase):
    name: Literal["Wondrous Trick: Miracle Parade"] = "Wondrous Trick: Miracle Parade"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        ret = super().get_actions(match)
        damage_action = ret[1]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        damage_action.create_objects += [
            self.create_summon("Grin-Malkin Hat"),
            self.create_character_status("Prop Surplus"),
        ]
        return ret


class ConclusiveOvation_4_3(SkillTalent):
    name: Literal["Conclusive Ovation"]
    character_name: Literal["Lyney"] = "Lyney"
    version: Literal["4.3"] = "4.3"
    skill: Literal["Prop Arrow"] = "Prop Arrow"
    usage: int = 1
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        """
        Reset usage
        """
        self.usage = 1
        return []

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Match,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return value
        if self.usage == 0:
            # no usage left
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self
            return value
        if value.target_position.player_idx == self.position.player_idx:
            # self target self
            return value
        target: CharacterBase = match.get_object(value.target_position)  # type: ignore
        assert target is not None
        if ElementType.PYRO not in target.element_application:
            # target have no pyro
            return value
        source = match.get_object(value.position)
        if source is not None and source.type == ObjectType.SKILL:
            # is skill, find corresponding character
            source = match.player_tables[value.position.player_idx].characters[
                value.position.character_idx
            ]
        if source is None or source.name not in ["Lyney", "Grin-Malkin Hat"]:
            # not from Lyney or Grin-Malkin Hat
            return value
        # increase damage by 2
        assert mode == "REAL"
        self.usage -= 1
        value.damage += 2
        return value


class Lyney_4_3(CharacterBase):
    name: Literal["Lyney"]
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | PropArrow
        | ElementalSkillBase
        | WondrousTrickMiracleParade
    ] = []
    faction: List[FactionType] = [
        FactionType.FONTAINE,
        FactionType.FATUI,
        FactionType.ARKHE_PNEUMA,
    ]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Card Force Translocation",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            PropArrow(),
            ElementalSkillBase(
                name="Bewildering Lights",
                damage_type=DamageElementalType.PYRO,
                cost=ElementalSkillBase.get_cost(self.element),
            ),
            WondrousTrickMiracleParade(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Lyney": {
        "names": {"en-US": "Lyney", "zh-CN": "林尼"},
        "image_path": "cardface/Char_Avatar_Liney.png",  # noqa: E501
        "id": 1310,
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
    },
    "SKILL_Lyney_NORMAL_ATTACK/Card Force Translocation": {
        "names": {"en-US": "Card Force Translocation", "zh-CN": "迫牌易位式"},
        "descs": {
            "4.3": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Lyney_NORMAL_ATTACK/Prop Arrow": {
        "names": {"en-US": "Prop Arrow", "zh-CN": "隐具魔术箭"},
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Pyro DMG and summons Grin-Malkin Hat.\nHowever, if this character's HP is higher than the opposing active character's, they will gain 1 stack of Prop Surplus, and then take 1 Piercing DMG.",  # noqa: E501
                "zh-CN": "造成2点火元素伤害，召唤怪笑猫猫帽，累积1层隐具余数。\n如果本角色生命值至少为6，则对自身造成1点穿透伤害。",  # noqa: E501
            }
        },
    },
    "SUMMON/Grin-Malkin Hat": {
        "names": {"en-US": "Grin-Malkin Hat", "zh-CN": "怪笑猫猫帽"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deal 1 Pyro DMG.\nUsage(s): 1 (Can stack. Max 2 stacks.)",  # noqa: E501
                "zh-CN": "结束阶段：造成1点火元素伤害。\n可用次数：1（可叠加，最多叠加到2次）",  # noqa: E501
            }
        },
        "image_path": "cardface/Summon_Liney.png",  # noqa: E501
    },
    "CHARACTER_STATUS/Prop Surplus": {
        "names": {"en-US": "Prop Surplus", "zh-CN": "隐具余数"},
        "descs": {
            "4.3": {
                "en-US": "Prop Surplus can stack up to 3 times.\nWhen this character uses Bewildering Lights: +1 DMG for each Prop Surplus stack. After the skill is finalized, consume all Prop Surplus, healing 1 HP for this character per stack.",  # noqa: E501
                "zh-CN": "隐具余数最多可以叠加到3层。\n角色使用眩惑光戏法时：每层隐具余数使伤害+1。技能结算后，耗尽隐具余数，每层治疗角色1点。",  # noqa: E501
            }
        },
    },
    "SKILL_Lyney_ELEMENTAL_SKILL/Bewildering Lights": {
        "names": {"en-US": "Bewildering Lights", "zh-CN": "眩惑光戏法"},
        "descs": {
            "4.3": {"en-US": "Deals 3 Pyro DMG.", "zh-CN": "造成3点火元素伤害。"}
        },
    },
    "SKILL_Lyney_ELEMENTAL_BURST/Wondrous Trick: Miracle Parade": {
        "names": {
            "en-US": "Wondrous Trick: Miracle Parade",
            "zh-CN": "大魔术·灵迹巡游",
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Pyro DMG, summons 1 Grin-Malkin Hat, adds 1 stack of Prop Surplus.",  # noqa: E501
                "zh-CN": "造成3点火元素伤害，召唤怪笑猫猫帽，累积1层隐具余数。",  # noqa: E501
            }
        },
    },
    "TALENT_Lyney/Conclusive Ovation": {
        "names": {"en-US": "Conclusive Ovation", "zh-CN": "完场喝彩"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Lyney, equip this card.\nAfter Lyney equips this card, immediately use Prop Arrow.\nWhen Lyney is on the field and has this card equipped, the damage dealt by Lyney and your Grin-Malkin Hat will deal +1 DMG to characters affected by Pyro.\n(You must have Lyney in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为林尼时，装备此牌。\n林尼装备此牌后，立刻使用一次隐具魔术箭。\n装备有此牌的林尼在场时，林尼自身和怪笑猫猫帽对具有火元素附着的角色造成的伤害+2。（每回合1次）\n（牌组中包含林尼，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Liney.png",  # noqa: E501
        "id": 213101,
    },
}


register_class(
    GrinMalkinHat_4_3 | PropSurplus_4_3 | ConclusiveOvation_4_3 | Lyney_4_3, desc
)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....event import SkillEndEventArguments

from ....modifiable_values import DamageIncreaseValue

from ....action import CreateObjectAction

from ....consts import ObjectPositionType, SkillType, WeaponType

from ....struct import Cost, ObjectPosition
from .base import RoundEffectWeaponBase, WeaponBase


class AmosBow_3_7(RoundEffectWeaponBase):
    name: Literal["Amos' Bow"]
    cost: Cost = Cost(same_dice_number=3)
    version: Literal["3.7"] = "3.7"
    weapon_type: WeaponType = WeaponType.BOW
    max_usage_per_round: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        condition_satisfied = False
        if value.cost.total_dice_cost + value.cost.charge >= 5 and self.usage > 0:
            # have usage and costs satisfied
            self.damage_increase = 3
            condition_satisfied = True
        current_damage = value.damage
        super().value_modifier_DAMAGE_INCREASE(value, match, mode)
        if value.damage > current_damage:
            # value modified
            if condition_satisfied:
                # decrease usage
                assert mode == "REAL"
                self.usage -= 1
        self.damage_increase = 1
        return value


class ElegyForTheEnd_3_7(WeaponBase):
    name: Literal["Elegy for the End"]
    cost: Cost = Cost(same_dice_number=3)
    version: Literal["3.7"] = "3.7"
    weapon_type: WeaponType = WeaponType.BOW

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If equipped and self use elemental burst, create a status
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not equipped or not self use skill
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_BURST:
            # not elemental burst
            return []
        return [
            CreateObjectAction(
                object_name="Millennial Movement: Farewell Song",
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        ]


register_class(AmosBow_3_7 | ElegyForTheEnd_3_7)
